# This file is available at http://tu-dresden.de/zih/perf
#
# Copyright (c) 2014
# Technische Universitaet Dresden, Germany
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 2
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import os
import sys
import pprint
import collections
import csv
#import traceback
#from ctypes import *
import ctypes
import cProfile, pstats

sys.path.append(os.environ['PERF_EXEC_PATH'] + \
    '/scripts/python/Perf-Trace-Util/lib/Perf/Trace')
sys.path.append(os.environ['PERF_EXEC_PATH'] + \
    '/libexec/perf-core/scripts/python/Perf-Trace-Util/lib/Perf/Trace')

from perf_trace_context import *


from Core import *

from otf import *

inactive_cpu = -1

NSEC_IN_SEC = 1000000000

CLOCK_REALTIME = 0
CLOCK_MONOTONIC = 1 # see <linux/time.h>

class timespec(ctypes.Structure):
    _fields_ = [
        ('tv_sec', ctypes.c_long),
        ('tv_nsec', ctypes.c_long)
]

librt = ctypes.CDLL('librt.so.1', use_errno=True)
clock_gettime = librt.clock_gettime
clock_gettime.argtypes = [ctypes.c_int, ctypes.POINTER(timespec)]

def total_time(s, ns):
    return (s*1000000000 + ns)


def monotonic_time():
    t = timespec()
    if clock_gettime(CLOCK_MONOTONIC, ctypes.pointer(t)) != 0:
        errno_ = ctypes.get_errno()
        raise OSError(errno_, os.strerror(errno_))
    return total_time(t.tv_sec, t.tv_nsec) - int(os.getenv('DIFF_TIME', 0))

def global_time():
    t = timespec()
    if clock_gettime(CLOCK_REALTIME, ctypes.pointer(t)) != 0:
        errno_ = ctypes.get_errno()
        raise OSError(errno_, os.strerror(errno_))
    return total_time(t.tv_sec, t.tv_nsec)

def time_offset():
    return time_offset.value

time_offset.value = global_time() - monotonic_time()

def correct_time(t):
    return (t + time_offset())


##
# The configured debug level. Use 0 for errors only, 1 for warnings 
# and 2 for additional information and 3 for debug information.
##
debug = int(os.getenv('PERF_OTF_VERBOSE', '1'))

##
# Whether or not to reuse ended processes 
# Reused processes do not have a PID in their name
# but instead carry markers to signal the beginning
# of processes. 
##
reuse = os.getenv('PERF_OTF_REUSEPROCS', 'false').lower() in ('true', 'yes', 'y', '1')

print "Reuse processes: {}".format(reuse)

def dprint(l, s):
    global debug
    if debug >= l:
        print s

# External or hardcoded config
def config_group():
    """Returns grouping configuration, either 'THREAD' or 'CORE'"""
    return config_group.value

def config_namestub():
    """Returns otf namestub"""
    return config_namestub.value

def config_process_file():
    return config_process_file.value

def config_counter_stackdepth():
    return config_counter_stackdepth.value
      
def config_counter_pid():
    return config_counter_pid.value
      
def config_counter_tid():
    return config_counter_tid.value
      
config_profile                  = False
config_group.value              = intern(os.getenv('PERF_OTF_GROUP', 'THREAD').upper())
config_namestub.value           = intern(os.getenv('PERF_OTF_NAMESTUB', 'out'))
config_process_file.value       = intern(os.getenv('PERF_OTF_PROCESS_FILE', 'ps.ssv'))
config_counter_stackdepth.value = False
config_counter_pid.value        = False
config_counter_tid.value        = False

# Internal config updated during reading the trace
# TODO read that from header instead of guessing from the first occurance
def config_has_sched_switch(val = None):
    if (not val is None):
        config_has_sched_switch.value = val
    return config_has_sched_switch.value

config_has_sched_switch.value = False

def config_has_sched_fork(val = None):
    if (not val is None):
        config_has_sched_fork.value = val
    return config_has_sched_fork.value

config_has_sched_fork.value = False

def config_has_sched_exit(val = None):
    if (not val is None):
        config_has_sched_fork.value = val
    return config_has_sched_fork.value

config_has_sched_exit.value = False


class ProcessSnapshotEntry:
    def __init__(self, row):
        try:
            self.pid = int(row['PID'])
            self.tid = int(row['TID'])
            self.ppid = int(row['PPID'])
            self.command = row['COMMAND']
        except:
            dprint(0, "Failed to handle ProcessSnapshot row:")
            pprint.pprint(row)
            raise
    
    def __str__(self):
        return "PSE " + self.__dict__.__str__()


def get_process_dict():
    """Returns dict of tid -> 
       {"PID" : ?,"TID" : ?, "PPID" : ?, "COMM" : ?, "COMMAND" : ?}"""
    if (get_process_dict.value is None):
        pdict = {}
        try:
            with open(config_process_file(), 'rb') as file:
                reader = csv.DictReader(file, delimiter=';')
                for row in reader:
                    entry = ProcessSnapshotEntry(row)
                    key = entry.tid
                    assert not key in pdict, "Duplicate tid in process file"
                    pdict[key] = entry
        except IOError:
            dprint(1, "Warning: could not read ps file: {}".format(config_process_file()))
        get_process_dict.value = pdict
    return get_process_dict.value

get_process_dict.value = None

####
# Process 
####
class Process(object):
    count = 0
    map = {}
    list = []
    def __init__(self, key, parent, time):
        Process.count = Process.count + 1
        self.id = Process.count
        self.stack = []
        self.parent = parent
        self.time = time
        Process.map[key] = self
        Process.list.append(self)
        self.begin(time)
        self.children = list()
        if not parent is None:
            parent.addChild(self)

    def writeDefProcess(self):
        if (self.parent is None):
            parent_id = 0
        else:
            parent_id = self.parent.id
        
        ret = OTF_Writer_writeEndProcess(writer, self.time, self.id)
        assert(ret == 1), "Failed to write EndProcess record"
        ret = OTF_Writer_writeDefProcess(writer, 0, self.id, self.name, parent_id)
        assert(ret == 1), "Failed to write DefProcess record"
    
    def begin(self, time):
        ret = OTF_Writer_writeBeginProcess(writer, time, self.id)
        assert(ret == 1), "Failed to write BeginProcess record"
        self.alive = True
        self.start = time
        if config_counter_stackdepth():
            counters["stackdepth"].write(time, self, 0)
    
    def end(self, time):
        if self.alive:
            self.leaveAll(time)
            self.timeOfDeath = time
            self.time = time
            self.setCpu(inactive_cpu, time)
            self.alive = False
            #OTF_Writer_writeEndProcess(writer, time, self.id)
        
    def enter(self, time, function):
        if (not self.alive):
            zombie_time = time - self.timeOfDeath
            # Grace period of 1 ms for zombies to do something
            if zombie_time > 1000000:
                dprint(1, "Trying to enter function {} on {} that already died {}ns ago.".format(function.name, self.name, zombie_time))
            return
        self.stack.append(function)
        self.time = time
        assert (time >= self.time), "Enter is earlier than last event"
        ret = OTF_Writer_writeEnter(writer, time, function.id, self.id, 0)
        assert(ret == 1), "Failed to write Enter record"

    def leave(self, time, function = None):
        assert self.stack, "Leaving an empty stack is not allowed..."
        assert self.alive, "Why am I dead :("
        
        self.time = time
        
        stackfun = self.stack.pop()

        assert (time >= self.time), "Leave is earlier than last enter"
        t = time
        
        assert (function is None or function is stackfun), "Unbalanced stack"
        ret = OTF_Writer_writeLeave(writer, long(t), stackfun.id, self.id, 0)
        assert(ret == 1), "Failed to write Leave record"
        
    def leaveAll(self, time):
        while (self.stack):
            self.leave(time)

    def setSampleFreq(self, freq):
        self.sample_freq = freq
        self.sample_period = (1000000000.0 / freq)
        
    def addChild(self, child):
        self.children.append(child)

class CoreProcess(Process):
    def __init__(self, core, time):
        key = core
        self.cpu = core
        parent = None
        super(CoreProcess, self).__init__(key, parent, time)
    
    @property
    def name(self):
        return "Core {}".format(self.cpu)

    @staticmethod
    def ensure(core, time):
        # TODO optimize to avoid quadratic complexity
        if (core in Process.map):
            return Process.map[core]
        for smallcore in range(0, core):
            if (not smallcore in Process.map):
                CoreProcess(smallcore, time)
        return CoreProcess(core, time)

class ThreadProcess(Process):
    def __init__(self, pid, tid, ppid, time, cpu=inactive_cpu):
        key = tid
        self.pid = pid
        self.tid = tid
        self.cpu = None
        self.reused = False
        self.orphan = True
        if (pid != tid):
            ppid = pid # Overwriting parent process with main thread in process...
        if (not ppid is None):
            pkey = ppid
            assert pkey != key, "I (%s) cannot be my own parent (%s) process" % (key, pkey)
            parent = ThreadProcess.ensure(ppid, ppid, time)
        else:
            parent = None
        super(ThreadProcess, self).__init__(key, parent, time)
        self.setCpu(cpu, time)
        
    def setCommand(self, command, orphan = False):
        self.command = command.split(' -')[0] # split command line arguments
        self.orphan = orphan
        
    def setOrphan(self):
        self.setCommand("(orphan)", True)
        
    @property
    def name(self):
        return self.getName()
    
    def getName(self, withpid=False):
        name = ""
        if (not self.reused) or (withpid):
            if (self.pid != self.tid):
                name = "[{},{}]".format(self.pid, self.tid)
            else:
                name = "[{}]".format(self.pid)
            name += ": " 
        try:
            name += self.command
        except AttributeError: # XXX FIXME is that really necessary
            None
        return name
        
    @property
    def basename(self):
        bn = ThreadProcess.getBasename(self.command)
        #print "Basename('{}') => '{}'".format(self.command, bn)
        return bn
        
    @staticmethod
    def getBasename(command):
        return command.split(' -')[0].split('/')[-1]    
    
    @staticmethod
    def ensure(pid, tid, time, cpu=inactive_cpu):
        key = tid
        if (key in Process.map):
            return Process.map[key]
        pdict = get_process_dict()
        if (key in pdict):
            row = pdict[key]
            proc = ThreadProcess(row.pid, row.tid, row.ppid, time, cpu)
            proc.setCommand(row.command)
        else:
            proc = ThreadProcess(pid, tid, None, time, cpu)
            if (pid == 0): # of course we cannot find process 0 (aka sched/swapper)
                proc.setCommand("swapper")
            else:
                dprint(1, "Failed to find process (%s, %s) in process table, allocating orphan" % (pid, tid))
                proc.setOrphan()
        
        return Process.map[key]
    
    threadByCpu = {}
    def setCpu(self, cpu, time):
        # don't set the CPU for dead processes
        if not self.alive:
            return
        if (not config_has_sched_switch()) and (cpu != inactive_cpu) and (cpu in ThreadProcess.threadByCpu):
            rival_thread = ThreadProcess.threadByCpu[cpu]
            if (not rival_thread is self):
                #print "Kicking {} in favor of {} from cpu {} at t={}".format(rival_thread.name, self.name, cpu, time)
                # Kick him, this is now our cpu
                idle_fun = get_function("idle", "idle")
                idle_fun.isIdle = True
                rival_thread.leaveAll(time)
                rival_thread.setCpu(inactive_cpu, time)
                rival_thread.enter(time, idle_fun)
        
        ThreadProcess.threadByCpu[cpu] = self

#        if cpu == -1:
#            #print "setCpu called with -1: "
#            #traceback.print_stack()
#            #exit(1)
#            pass
        if (self.cpu != cpu):
            self.cpu = cpu
            # We write the cpu + 1 to distinguish INACTIVE(-1), UNDEFINED(0), CPU(1+)
            if (cpu != inactive_cpu):
                cpu += 1
            counters["sched"].write(time, self, cpu)
            
    def revive(self, cpu, pid, tid, time):
    
        if not self.reused:
            # write a marker for the beginning of the original process
#            ret = OTF_Writer_writeMarker(writer, self.start, self.id, 1000, self.getName(True))
#            assert(ret == 1), "Failed to write Marker"
            markers[self.id] = ProcessCommMarker("ProcessChange", OTF_MARKER_TYPE_HINT, self)
            #markers[self.id].write(self.start, self.getName(True))
            self.reused = True
        
        self.alive = True
        self.pid = pid
        self.tid = tid
        Process.map[tid] = self
        self.setCpu(cpu, time)
        # make this change visible
        markers[self.id].write(time, self.getName(True))
#        ret = OTF_Writer_writeMarker(writer, time, self.id, 1000, self.getName(True))     
#        assert(ret == 1), "Failed to write Marker"

min_time = 2 ** 62
max_time = 0

def get_process(core, pid, tid, time):
    global min_time
    global max_time
    if (time < min_time):
        min_time = time
    if (time > max_time):
        max_time = time

    if ('THREAD' == config_group()):
        proc = ThreadProcess.ensure(pid, tid, time, core)
    elif ('CORE' == config_group()):
        proc = CoreProcess.ensure(core, time)
    else:
        raise Exception('Invalid group', config_group())

    if (config_counter_pid()):
        counters['pid'].write(time, proc, pid)
    if (config_counter_tid()):
        counters['tid'].write(time, proc, tid)

    return proc
    
def get_process_by_name(core, pid, tid, ppid, name, time):
    global min_time
    global max_time
    if (time < min_time):
        min_time = time
    if (time > max_time):
        max_time = time
        
    proc = None
    
    if ('CORE' == config_group()):
        proc = CoreProcess.ensure(core, time)
    elif ('THREAD' == config_group()):
        # look for a process we can reuse
        if reuse and (ppid in Process.map):
            print "Trying to find a process to reuse for {}".format(ThreadProcess.getBasename(name))
            parent = Process.map[ppid]
            for child in parent.children:
                if not child.alive and ThreadProcess.getBasename(name) == child.basename:
                    # we can reuse this process
                    proc = child
                    proc.revive(core, pid, tid, time)
                    print "Reusing process {} ({})".format(proc.id, proc.name)
                elif not child.alive:
                    print "{} != {}".format(ThreadProcess.getBasename(name), child.basename)
        if proc is None:
            proc = ThreadProcess(pid, tid, ppid, time)
        
    return get_process(core, pid, tid, time)
    
####
# FunctionGroup 
####
class FunctionGroup:
    count = 0
    map = {}
    def __init__(self, name):
        FunctionGroup.count = FunctionGroup.count + 1
        self.id = FunctionGroup.count
        self.name = name
        self.source = 0
        self.functions = {}
        OTF_Writer_writeDefFunctionGroup(writer, 0, self.id, self.name)

def get_functionGroup(name):
    if (not name in FunctionGroup.map):
        FunctionGroup.map[name] = FunctionGroup(name)
    return FunctionGroup.map[name]

####
# Function 
####
class Function:
    count = 0
    #map = {}
    def __init__(self, name, group):
        Function.count = Function.count + 1
        self.id = Function.count
        self.name = name
        self.group = get_functionGroup(group)
        self.source = 0
        self.isIdle = False
        OTF_Writer_writeDefFunction(writer, 0, self.id, self.name, self.group.id, self.source)

    def enter(self, time, process):
        process.enter(time, self)

    def leave(self, time, process):
        process.leave(time, self)

def get_function(name, group):
    if (not group in FunctionGroup.map):
        FunctionGroup.map[group] = FunctionGroup(group)
    g = FunctionGroup.map[group]
    if (not name in g.functions):
        g.functions[name] = Function(name, group)
    return g.functions[name]

####
# Counter 
####
def Counter(name, properties, unit):
    if (OTF_COUNTER_VARTYPE_ISINTEGER(properties)):
        if (OTF_COUNTER_VARTYPE_ISUNSIGNED(properties)):
            return UnsignedCounter(name, properties, unit)
        elif (OTF_COUNTER_VARTYPE_ISSIGNED(properties)):
            return SignedCounter(name, properties, unit)
        else:
            assert false, "Invalid OTF counter properties"
    else:
        return DoubleCounter(name, properties, unit)

class BaseCounter:
    count = 0
    def __init__(self, name, properties, unit):
        BaseCounter.count += 1
        self.id = BaseCounter.count
        self.properties = properties
        OTF_Writer_writeDefCounter(writer, 0, self.id, name, properties, 1000, unit)

class UnsignedCounter(BaseCounter):
    def __init__(self, name, properties, unit):
        assert OTF_COUNTER_VARTYPE_ISUNSIGNED(properties), "Inconsistent properties"
        BaseCounter.__init__(self, name, properties, unit)
    
    def write(self, time, process, value):
        if process.alive:
            otf_value = OTF_Unsigned2Counter(value)
            OTF_Writer_writeCounter(writer, time, process.id, self.id, otf_value)
            process.time = time

class SignedCounter(BaseCounter):
    def __init__(self, name, properties, unit):
        assert OTF_COUNTER_VARTYPE_ISSIGNED(properties), "Inconsistent properties"
        BaseCounter.__init__(self, name, properties, unit)
    
    def write(self, time, process, value):
        if process.alive:
            otf_value = OTF_Signed2Counter(value)
            OTF_Writer_writeCounter(writer, time, process.id, self.id, otf_value)
            process.time = time

class DoubleCounter(BaseCounter):
    def __init__(self, name, properties, unit):
        assert OTF_COUNTER_VARTYPE_ISDOUBLE(properties), "Inconsistent properties"
        BaseCounter.__init__(self, name, properties, unit)
    
    def write(self, time, process, value):
        if process.alive:
            otf_value = OTF_Double2Counter(value)
            OTF_Writer_writeCounter(writer, time, process.id, self.id, otf_value)
            process.time = time

# TODO Evaluate if Signed/Unsigned is better here
class SignedIncrementCounter(SignedCounter):
    def __init__(self, name, properties, unit):
        SignedCounter.__init__(self, name, properties, unit)
        self.values = {}

    def add(self, time, process, value):
        if process.id in self.values:
            self.values[process.id] += value
        else:
            self.values[process.id] = 0
        self.write(time, process, self.values[process.id])

# TODO Evaluate if Signed/Unsigned is better here
class UnsignedIncrementCounter(UnsignedCounter):
    def __init__(self, name, properties, unit):
        UnsignedCounter.__init__(self, name, properties, unit)
        self.values = {}

    def add(self, time, process, value):
        if process.id in self.values:
            self.values[process.id] += value
        else:
            self.values[process.id] = 0
        self.write(time, process, self.values[process.id]);

class Object:
    pass

#counters = Object()
counters = dict()

####
# Marker 
####
class Marker:
    count = 0
    def __init__(self, name, mtype):
        Marker.count += 1
        self.id = Marker.count
        ret = OTF_Writer_writeDefMarker(writer, 0, self.id, name, mtype)
        assert(ret == 1), "Failed to write Marker Definition record"
    
    def write(self, time, process, text):
        ret = OTF_Writer_writeMarker(writer, time, process, self.id, text)
        assert(ret == 1), "Failed to write Marker"
        
class ProcessMarker(Marker):
    def __init__(self, name, mtype, process):
        Marker.__init__(self, "{} ({})".format(name, process.id), mtype)
        self.pid = process.id
        
    def write(self, time, text):
        ret = OTF_Writer_writeMarker(writer, time, self.pid, self.id, text)
        assert(ret == 1), "Failed to write Marker"
        
class ProcessCommMarker(ProcessMarker):
    def __init__(self, name, mtype, process):
        #ProcessMarker.__init__(self, name, mtype, process)
        self.pid = process.id
        pass
        
    def write(self, time, text):
        #super(ProcessCommMarker, self).write(time, text)
        # put a short communication here
        OTF_Writer_writeSendMsg(writer, time, self.pid, self.pid, 0, 0, 0, 0)
        OTF_Writer_writeRecvMsg(writer, time, self.pid, self.pid, 0, 0, 0, 0)


markers = dict()

def otf_init(namestub):
    otf_init.manager = OTF_FileManager_open(100)
    global writer
    writer = OTF_Writer_open(namestub, 1000000, otf_init.manager)
    OTF_Writer_setCompression(writer, 8)

    OTF_Writer_writeDefCreator(writer, 0, "perf-otf.py v0.1.4")
    OTF_Writer_writeDefTimerResolution(writer, 0, 1000000000)

    OTF_Writer_writeDefCounterGroup(writer, 0, 1000, "status counters")
    
    global counters
    if config_counter_pid():
        counters["pid"]     = Counter("Process ID", OTF_COUNTER_TYPE_ABS|OTF_COUNTER_SCOPE_NEXT|OTF_COUNTER_VARTYPE_UNSIGNED8,  "#")
    if config_counter_tid():
        counters["tid"]     = Counter("Thread ID", OTF_COUNTER_TYPE_ABS|OTF_COUNTER_SCOPE_NEXT|OTF_COUNTER_VARTYPE_UNSIGNED8,  "#")
        
    counters["cpufreq"] = Counter("CPU Frequency", OTF_COUNTER_TYPE_ABS|OTF_COUNTER_SCOPE_NEXT|OTF_COUNTER_VARTYPE_UNSIGNED8,  "MHz")
    counters["sched"]   = Counter("Scheduled on CPU", OTF_COUNTER_TYPE_ABS|OTF_COUNTER_SCOPE_NEXT|OTF_COUNTER_VARTYPE_SIGNED8,  "#")
    if config_counter_stackdepth():
        counters["stackdepth"] = Counter("Stackdepth", OTF_COUNTER_TYPE_ABS|OTF_COUNTER_SCOPE_NEXT|OTF_COUNTER_VARTYPE_UNSIGNED4, "#")
    counters["pages"]   = SignedIncrementCounter("Pages allocated", OTF_COUNTER_TYPE_ABS|OTF_COUNTER_SCOPE_NEXT|OTF_COUNTER_VARTYPE_SIGNED8, "#")

def otf_finalize():
    nprocs = len(Process.map)

#    sortedProcs = collections.OrderedDict(sorted(Process.map.items()))
#    for key, process in sortedProcs.iteritems():
    for process in Process.list:
        process.writeDefProcess()

    OTF_Writer_writeDefProcessGroup(writer, 0, 100000, "all", nprocs, range(1, nprocs + 1))

    OTF_Writer_writeDefTimeRange(writer, 0, min_time, max_time, None)

    OTF_Writer_close(writer)
    OTF_FileManager_close(otf_init.manager)

num_samples       = 0
num_bogus_samples = 0

def trace_begin():
    if config_profile:
        global pr
        pr = cProfile.Profile()
        pr.enable()
    dprint(0, "in trace_begin")
    otf_init(config_namestub())

def trace_end():
    global pr
    dprint(0, "in trace_end")
    if num_samples > 0:
        pc = float(num_bogus_samples) / float(num_samples) * 100
    else:
        pc = "n/a"
    dprint(0, "Bogus samples: {} out of {} ({}%)".format(num_bogus_samples, num_samples, pc))

    otf_finalize()
    if config_profile:
        pr.disable()
        st = pstats.Stats(pr)
        st.dump_stats("perf2otf.pyprofile")
        st = st.sort_stats("time")
        st.print_stats()
        
def process_callchain(time, process, callchain):
    
    process.leaveAll(time)
    # store callchain for use when the process is switched back in
    process.callchain = callchain

    callchain.reverse()
    for csentry in callchain:
        name = "Unknown"
        if ("sym" in csentry):
            name = csentry["sym"]["name"]
        
        if ("dso" in csentry):
            group = csentry["dso"]
            function = get_function(name, group)
            process.enter(time, function)
        
    if config_counter_stackdepth():
        counters["stackdepth"].write(time, process, len(callstack))

def process_event(param_dict):
#    print "Process_event:"
#    pprint.pprint(param_dict);
    event_attr = param_dict["attr"]
    sample     = param_dict["sample"]
    raw_buf    = param_dict["raw_buf"]
    comm       = param_dict["comm"]
    ev_name    = param_dict["ev_name"]
    time       = correct_time(sample["time"])
        
    global num_samples
    global num_bogus_samples

    num_samples += 1

    dprint(4, "SAMPLE: '{}' on CPU {} @{} from '{}'".format(ev_name, sample["cpu"], time, comm))

    assert time >= 0, 'negative time (%d) is not allowed' % time

    process = get_process(sample["cpu"], sample["pid"], sample["tid"], time)
    
    # Set the command if we have no information about it so far
    # e.g., due to missing sched events
    if 'THREAD' == config_group() and process.orphan:
        process.setCommand(comm)


    # Symbol and dso info are not always resolved
    if ("dso" in param_dict):
        dso = param_dict["dso"]
    else:
        dso = "Unknown_dso"

    if (param_dict.has_key("symbol")):
        symbol = param_dict["symbol"]
    else:
        symbol = "Unknown_symbol"

    # If we do not record with -a, the cpu will be invalid indicated by 4294967295
    if sample["cpu"] != 4294967295 and process.cpu != sample["cpu"]:
        # If we have sched_switch events, then we assume they are always right and samples
        # may be a little incorrect
        if config_has_sched_switch() and sample["pid"] != 0:
            num_bogus_samples += 1
            dprint(2, "Discarding bogus sample of type '{}' @{} ({}:{}, pid:{}, tid:{}), cpus don't match ({}!={})".format(
                ev_name, time, dso, symbol, sample['pid'], sample['tid'], process.cpu, sample["cpu"]))
            return
        else:
            process.setCpu(sample['cpu'], time)

    # TODO: Should this maybe be configurable?
    if (ev_name == "cycles" or ev_name == "cpu-clock"):
        process_callchain(time, process, param_dict["callchain"]);    
        
    if (ev_name not in counters):
        counters[ev_name] = UnsignedIncrementCounter(ev_name, OTF_COUNTER_TYPE_ACC|OTF_COUNTER_SCOPE_LAST|OTF_COUNTER_VARTYPE_UNSIGNED8,  "#")
    counters[ev_name].add(time, process, sample['period'])
        
        
##
# Handler for perf internal fork/exit events (always recorded)
##
def process_fork(event_name, context, common_cpu, common_secs, common_nsecs, common_pid,
        time, pid, tid, ppid, ptid):
    dprint(3, "{}: pid={}, tid={}, ppid={}, ptid={}, time={}".format(event_name, pid, tid, ppid, ptid, time))
    if (not config_has_sched_fork()) and ('THREAD' == config_group()):
        time = correct_time(total_time(common_secs, common_nsecs))
        proc = ThreadProcess(pid, tid, ppid, time)
        pproc = get_process(common_cpu, ppid, ptid, time)
        
def process_exit(event_name, context, common_cpu, common_secs, common_nsecs, common_pid,
        time, pid, tid, ppid, ptid):
    dprint(3, "{}: pid={}, tid={}, ppid={}, ptid={}, time={}".format(event_name, pid, tid, ppid, ptid, time))
    
    if (not config_has_sched_exit()) and ('THREAD' == config_group()):
        if tid in Process.map:
            time = correct_time(total_time(common_secs, common_nsecs))
            key = tid
            Process.map[key].end(time)

##
# Handler for various subsystem events.
# If an event is missing, feel free to add it below. 
##

def power__cpu_frequency(event_name, context, common_cpu, common_secs, common_nsecs, common_pid, common_comm, common_callchain, state, cpu_id):
    time = correct_time(total_time(common_secs, common_nsecs))
    process = get_process(cpu_id, common_pid, common_pid, time)
    process_callchain(time, process, common_callchain);    
    counters["cpufreq"].write(time, process, state)

# Old deprecated version
def power__power_frequency(event_name, context, common_cpu, common_secs, common_nsecs, common_pid, common_comm, common_callchain, state, cpu_id):
    power__cpu_frequency(event_name, context, common_cpu, common_secs, common_nsecs, common_pid, common_comm, state, cpu_id, callchain)

def sched__sched_process_exec(event_name, context, common_cpu,
    common_secs, common_nsecs, common_pid, common_comm, common_callchain,
    filename, pid, old_pid):
    dprint(3, "{}: pid={}, filename={}, common_comm={}, time={}".format(event_name, common_pid, filename, common_comm, common_secs * 1000000000 + common_nsecs))
    if ('THREAD' == config_group()):
        time = correct_time(total_time(common_secs, common_nsecs))
        key = pid
        Process.map[key].setCommand(filename)

def sched__sched_process_fork(event_name, context, common_cpu,
    common_secs, common_nsecs, common_pid, common_comm, common_callchain,
    parent_comm, parent_pid, child_comm, child_pid):
    dprint(3, "{}: pid={}, parent_pid={}, child_pid={}, child_comm={}, time={}".format(event_name, common_pid, parent_pid, child_pid, child_comm, common_secs * 1000000000 + common_nsecs))
    config_has_sched_fork(True)
    if ('THREAD' == config_group()):
        time = correct_time(total_time(common_secs, common_nsecs))
        #proc = ThreadProcess(child_pid, child_pid, parent_pid, time)
        proc = get_process_by_name(common_cpu, child_pid, child_pid, parent_pid, child_comm, time)
        proc.setCommand(child_comm) # for now until we get an exec (not happening for threads)
        process_callchain(time, proc, common_callchain);    
        #print "sched_process_fork: pid={}, ppid={}, child_comm={}, parent_comm={}".format(child_pid, parent_pid, child_comm, parent_comm)

def sched__sched_process_exit(event_name, context, common_cpu,
    common_secs, common_nsecs, common_pid, common_comm, common_callchain,
    comm, pid, prio):
    config_has_sched_exit(True)
    if ('THREAD' == config_group()):
        time = correct_time(total_time(common_secs, common_nsecs))
        key = pid
        Process.map[key].end(time)

def state2text(state):
    # TODO: document the meaning of the states!
    max = 512
    text = ""
    if (state & 1 != 0): text += "S"
    if (state & 2 != 0): text += "D"
    if (state & 4 != 0): text += "T"
    if (state & 8 != 0): text += "t"
    if (state & 16 != 0): text += "Z"
    if (state & 32 != 0): text += "X"
    if (state & 64 != 0): text += "x"
    # according to http://linux.about.com/od/commands/l/blcmdl1_top.htm state "W" meant "swapped out" but was removed in 2.3.x kernel
    #if (state & 128 != 0): text += "W" 
    if (state & (max - 1) == 0): text += "R"
    if (state & max != 0): text += "+"
    return text

def handle_wakeup(event_name, context, common_cpu, common_secs,
    common_nsecs, common_pid, common_comm, callchain,
    comm, pid, prio, success, target_cpu):
    time = correct_time(total_time(common_secs, common_nsecs))
    proc = get_process(common_cpu, common_pid, common_pid, time)
    
    if 'THREAD' == config_group():
        proc.setCpu(target_cpu, time)
    
    process_callchain(time, proc, callchain);    

def sched__sched_wakeup(event_name, context, common_cpu, common_secs,
           common_nsecs, common_pid, common_comm, common_callchain,
           comm, pid, prio, success, target_cpu):
    dprint(3, "WAKEUP process {} from CPU {} to target {} @{}".format(common_pid, common_cpu, target_cpu, (common_secs * 1000000000 + common_nsecs)))
    handle_wakeup(event_name, context, common_cpu, common_secs,
        common_nsecs, common_pid, common_comm, common_callchain,
        comm, pid, prio, success, target_cpu)

def sched__sched_wakeup_new(event_name, context, common_cpu, common_secs,
           common_nsecs, common_pid, common_comm, common_callchain,
           comm, pid, prio, success, target_cpu):
    dprint(3, "WAKEUP_NEW process {} from CPU {} to target {} @{}".format(common_pid, common_cpu, target_cpu, (common_secs * 1000000000 + common_nsecs)))
    handle_wakeup(event_name, context, common_cpu, common_secs,
        common_nsecs, common_pid, common_comm, common_callchain,
        comm, pid, prio, success, target_cpu)
    

def sched__sched_switch(event_name, context, common_cpu,
    common_secs, common_nsecs, common_pid, common_comm, common_callchain,
    prev_comm, prev_pid, prev_prio, prev_state, 
    next_comm, next_pid, next_prio):
    config_has_sched_switch(True)
    time = correct_time(total_time(common_secs, common_nsecs))
    if 'THREAD' == config_group():

        dprint(3, "SWITCH process {} for {} on cpu {} @{}".format(prev_pid, next_pid, common_cpu, time))
        
        if (prev_pid != 0): # Ignore swapper when switching
            prev_proc = get_process(common_cpu, prev_pid, prev_pid, time)
            prev_proc.setCpu(inactive_cpu, time)
            prev_proc.leaveAll(time)
        
            txt = state2text(prev_state)
            if (prev_state & 64) or (prev_state & 32): # Terminated state, probably already dead
                if prev_proc.alive:
                    dprint(2, "Killing zombie process {}.".format(prev_proc.name))
                    prev_proc.end(time)
            else:
                idle_fun = get_function("idle({})".format(txt), "idle({})".format(txt))
                idle_fun.isIdle = True
                prev_proc.enter(time, idle_fun)
                
        if (next_pid != 0): # Ignore swapper again
            next_proc = get_process(common_cpu, next_pid, next_pid, time)
            next_proc.setCpu(common_cpu, time)
            
            if next_proc.stack:
                if next_proc.stack[-1].isIdle:
                    next_proc.leave(time)
                if next_proc.stack:
                    dprint(2, "Process {} with non-empty stack activated by scheduler...".format(next_proc.name))
                    pprint.pprint(next_proc.stack)

            #next_proc.enter(time, get_function(next_comm, next_comm))
            #process_callchain(time, next_proc, common_callchain);                    
            if hasattr(next_proc, 'callchain'):
                # us the previously stored callchain and remove kernel calls from it
                next_proc.callchain.reverse()
                process_callchain(time, next_proc, [csentry for csentry in next_proc.callchain if "dso" not in csentry or csentry["dso"] != "[kernel.kallsyms]"])
    else:
        # XXX FIXME TODO write handler for CORE processes
        proc = get_process(common_cpu, next_pid, next_pid, time)
        process_callchain(time, proc, common_callchain);

def sched__sched_migrate_task(event_name, context, common_cpu,
    common_secs, common_nsecs, common_pid, common_comm, common_callchain,
    comm, pid, prio, orig_cpu, dest_cpu):
    # We don't need to handle that when sched_switch events are present
    dprint(3, "MIGRATE process {} from cpu {} to cpu {} @{}".format(pid, orig_cpu, dest_cpu, common_secs*1000000000 + common_nsecs))
    pass

def kmem__mm_page_alloc(event_name, context, common_cpu, common_secs, common_nsecs, common_pid, common_comm, common_callchain,
    page, order, gfp_flags, migratetype):
    
    time = correct_time(total_time(common_secs, common_nsecs))
    proc = get_process(common_cpu, common_pid, common_pid, time)
    if 'THREAD' == config_group():
        counters["pages"].add(time, proc, 1)

    process_callchain(time, proc, common_callchain);    


def kmem__mm_page_free_direct(event_name, context, common_cpu, common_secs, common_nsecs, common_pid, common_comm, common_callchain,
    page, order, proc):

    time = correct_time(total_time(common_secs, common_nsecs))
    proc = get_process(common_cpu, common_pid, common_pid, time)
    if 'THREAD' == config_group():
        counters["pages"].add(time, proc, -1)
        
    process_callchain(time, proc, common_callchain);    

unhandled_events = list()

def trace_unhandled(event_name, context, event_fields_dict):
    global unhandled_events
    if not event_name in unhandled_events:
        unhandled_events.append(event_name)
        dprint(1, 'UNKNOWN ' + event_name + ' '.join(['%s=%s'%(k,str(v))for k,v in sorted(event_fields_dict.items())]))

def print_header(event_name, cpu, secs, nsecs, pid, comm):
    print "%-20s %5u %05u.%09u %8u %-20s " % \
    (event_name, cpu, secs, nsecs, pid, comm)
