#!/bin/bash

control_c()
# run if user hits control-c
{
	echo "Stopping perf..."
	kill -s SIGINT $PERF_PID
}

PS_SSV_FILE="ps.ssv.$$"

# Check if python modules are available
python -c "import otf" || { echo "OTF python bindings are not available. Please add the path to the OTF python bindings to your PYTHONPATH environment variable!" ; exit 1 ; }
python -c "import Core" || { echo "PERF Core python module not found! Please add the path to the PERF Core python module to your PYTHONPATH environment variable!" ; exit 1 ; }

[ -z "$PERF2OTF" ] && { echo "Missing PERF2OTF environment variable! Please set PERF2OTF so that it points to the perf2otf.py script!" ; exit 1 ; }

# Get a list of available perf events
EVENTS_AVAIL=($(perf list | grep event | grep -v Raw | grep -v "List of" | cut -d '[' -f 1 | tr OR ' ' | tr '\n' ' '))

# Start our perf recording
EVENTS=""
EVENTS_C=""

check_event()
{
	local e
	for e in "${EVENTS_AVAIL[@]}"; do [[ "$e" == "$1" ]] && return 0; done
	return 1
}

add_event()
{
	if ! check_event "$1" ; then echo "Warning: Event $1 not available on this system! Dropping event!"; return; fi
	EVENTS="-e $1 $EVENTS"
	EVENTS_C="$EVENTS_C $1"
}

#add_event "cycles"
add_event "cpu-clock"
#add_event "sched:sched_process_exec"
#add_event "sched:sched_process_fork"
#add_event "sched:sched_process_exit"
#add_event "sched:sched_migrate_task"
add_event "branch-instructions"
#add_event "LLC-load-misses"
add_event "sched:sched_wakeup"
add_event "sched:sched_wakeup_new"
add_event "sched:sched_switch"
add_event "kmem:mm_page_alloc"
add_event "kmem:mm_page_free_direct"
#add_event "kmem:mm_pagevec_free"
#add_event "power:cpu_frequency"

echo "Starting perf while recording the following events:$EVENTS_C"

TRACEID=$$
#PERF_DATA=/dev/shm/perf.data.$TRACEID
# Use a local default but allow PERF_DATA to be set by the user
: ${PERF_DATA:="perf.data.$TRACEID"}
perf record -o $PERF_DATA -R --call-graph fp $EVENTS "$@" &
PERF_PID=$!

# Wait for processes to start
sleep 0.1

# Record process information | replace delimiting spaces with semicolons | replace first COMMAND with COMM in header
ps -ALo "pid,tid,ppid,comm,args" \
| awk '{ ORS=""; print $1 ";" $2 ";" $3 ";" $4 ";" ; for(i=5;i<=NF;i++){ print  $i " "}; printf "\n"; }' \
| sed -e 's/COMMAND/COMM/' \
| sed -e 's/ *$//g' \
 > $PS_SSV_FILE

trap control_c SIGINT

while kill -0 $PERF_PID > /dev/null 2>&1
do
	wait $PERF_PID
done

if [ ! -f $PERF_DATA ]
then
	echo "PERF data file missing!"
	exit -1
fi

echo "Generating by-thread trace [$PERF_DATA]..."
mkdir -p trace
export PERF_OTF_PROCESS_FILE=$PS_SSV_FILE
export PERF_OTF_NAMESTUB=trace/by_thread.$TRACEID
export PERF_OTF_GROUP=thread
perf script -i $PERF_DATA -s $PERF2OTF

echo "Generating by-core trace [$PERF_DATA]..."
export PERF_OTF_NAMESTUB=trace/by_core.$TRACEID
export PERF_OTF_GROUP=core
perf script -i $PERF_DATA -s $PERF2OTF
